from os import listdir
from os.path import isfile, join
from AutoRPE.UtilsRPE.Functions import calls_to_routines, calls_to_functions, \
    is_call_to_function, is_keyword_argument, is_a_number, \
    find_called_routine, find_call_arguments, find_called_function, \
    match_arguments_and_interface, type_of_contents, retrieve_variable, in_which_block , \
    match_argument_types_and_interface
from AutoRPE.UtilsRPE.Classes import ReadSourceFile, Function, Interface
import warnings


def track_dependencies(path_to_model_sources, vault=None, routines=[], functions=[]):
    # Find files in sources
    all_sources = [f for f in listdir(path_to_model_sources) if
                   isfile(join(path_to_model_sources, f)) and f.count(".f90")]
    # Process Sources!
    total = len(all_sources)
    for index, filename in enumerate(all_sources):
        print("\rTracing dependencies %i/%i  %20s" % (index + 1, total, filename), end="")
        track_dependencies_in_file(join(path_to_model_sources, filename), vault, routines, functions)


def track_dependencies_in_file(filename, vault=None, routines=[], functions=[]):
    try:
        current_block = vault[filename.split("/")[-1].replace(".f90", "")].main()
    except AttributeError:
        warnings.warn("No module %s" % filename)
        return
    f = ReadSourceFile(filename)

    for line in f.lines:
        routine_calls = calls_to_routines(line)
        if routine_calls:
            for call in routine_calls:
                called_routine = find_called_routine(call)
                if called_routine == "date_and_time":
                    continue
                called_arguments = find_call_arguments(call)
                if not called_arguments:
                    continue
                procedure = vault.get_procedure(called_routine)
                
                # Interface matching  
                indices_to_ignore = []
                if isinstance(procedure, Interface):
                    argument_types = [type_of_contents(argument, routine=current_block, vault=vault) for argument in called_arguments]
                    keyword_arguments = [_x.split("=")[0].strip() if is_keyword_argument(_x) else False for _x in called_arguments]
                    # Checking if some of the arguments does not necessarily need to be bounded to the dummy
                    # because there's already an existing version within the same interface with a different precision
                    for _in, _type in enumerate(argument_types):
                       new_argument_types = argument_types[:]
                       if _type == "single":
                          new_argument_types[_in] = "double"
                       elif _type == "double":
                          new_argument_types[_in] = "single"
                       else:
                          continue
                       try:
                          pp = match_argument_types_and_interface(new_argument_types, procedure, keyword_arguments=keyword_arguments, current_block=current_block, vault=vault) 
                          indices_to_ignore.append(_in)
                       except AssertionError as err:
                          continue
                    # Classical match
                    procedure = match_argument_types_and_interface(argument_types, procedure, keyword_arguments=keyword_arguments, current_block=current_block, vault=vault)

                if procedure is None:
                    continue
                for index, argument in enumerate(called_arguments):
                    if index in indices_to_ignore:
                       continue
                    if type_of_contents(argument, routine=current_block, vault=vault) not in ["single", "double","rpe_var"]:
                       continue
                    try:
                        var = retrieve_variable(argument, current_block, vault=vault)
                    except AssertionError:
                        continue
                    if is_keyword_argument(argument):
                        dummy_var = procedure.get_variable_by_name(argument.split("=")[0].strip())
                    else:
                        dummy_var = procedure.get_variable_by_position(index)
                    var.same_as.append(dummy_var)
                    dummy_var.same_as.append(var)
                    routines.append(called_routine)

        function_calls = calls_to_functions(line, vault)
        if function_calls:
            for call in function_calls:
                called_function = find_called_function(call)
                called_arguments = find_call_arguments(call)
                if not called_arguments:
                    continue
                procedure = vault.get_procedure(called_function)
                indices_to_ignore = []
                if isinstance(procedure, Function):
                    pass
                elif isinstance(procedure, Interface):
                    argument_types = [type_of_contents(argument, routine=current_block, vault=vault) for argument in called_arguments]
                    keyword_arguments = [_x.split("=")[0].strip() if is_keyword_argument(_x) else False for _x in called_arguments]
                    # Checking if some of the arguments does not necessarily need to be bounded to the dummy
                    # because there's already an existing version within the same interface with a different precision
                    for _in, _type in enumerate(argument_types):
                       new_argument_types = argument_types[:]
                       if _type == "single":
                          new_argument_types[_in] = "double"
                       elif _type == "double":
                          new_argument_types[_in] = "single"
                       else:
                          continue
                       try:
                          pp = match_argument_types_and_interface(new_argument_types, procedure, keyword_arguments=keyword_arguments, current_block=current_block, vault=vault) 
                          indices_to_ignore.append(_in)
                       except AssertionError as err:
                          continue
                    # Classical match
                    procedure = match_argument_types_and_interface(argument_types, procedure, keyword_arguments=keyword_arguments, current_block=current_block, vault=vault)

                else:
                    warnings.warn(
                        "Function not found in vault database %s\nWill be treated as external." % called_function)
                if procedure is None:
                    continue
                for index, argument in enumerate(called_arguments):
                    if index in indices_to_ignore:
                       continue
                    if is_call_to_function(argument) or is_a_number(argument):
                        continue
                    if type_of_contents(argument, routine=current_block, vault=vault) not in ["single", "double",
                                                                                              "rpe_var"]:
                        continue
                    try:
                        var = retrieve_variable(argument, current_block, vault=vault)
                    except AssertionError:
                        continue
                    dummy_var = procedure.get_variable_by_position(index)
                    var.same_as.append(dummy_var)
                    dummy_var.same_as.append(var)
                    functions.append(called_function)
        # If not, look if the code is entering or exiting a routine block
        current_block = in_which_block(line, current_block, vault=vault)


def propagate_dependencies(vault):
    for var in vault.variables:
        var.group_name = None

    variables = [variable for variable in vault.variables]

    groups = {}

    for var in variables:
        groups_found = []
        for v in var.same_as:
            if v.group_name is not None:
                groups_found.append(v.group_name)
        if not groups_found:
            groups[var_identifier(var)] = [var]
            var.group_name = var_identifier(var)
        else:
            selected_group = groups_found[0]
            var.group_name = selected_group
            groups[selected_group].append(var)
            for group in groups_found:
                if group != selected_group:
                    while len(groups[group]):
                        v = groups[group].pop(0)
                        v.group_name = selected_group
                        groups[selected_group].append(v)
                    # del groups[group]

    for group in groups.values():
        for var in group:
            var.same_as = [v for v in group if v != var]
            var.same_as = list(set(var.same_as))


def var_identifier(var):
    return "%s%s%s" % (var.name, var.routine.name, var.routine.module.name)
